//===========================================================================
// Dateiname : minidmx_mod.c
// Funktion  : DMX-Ausgang fr den PC
// Compiler  : WinAVR v20030312 (http://sourceforge.net/projects/winavr/)
// MCU       : AT90S2313
// Quarz     : 16000 MHz
// nderungen: 18.02.2001 - V1.0
//             - Erste ffentliche Version
//             05.02.2003 - V1.1
//             - LED1 zeigt nun ein erfolgreiches Senden der DMX Daten an
//             - LED2 zeigt Fehler beim Datenempfang vom PC an
//             - LED3 zeigt das Ende das Startvorgangs an (Daten knnen nun
//               gesendet werden)
//             21.03.2003 - V2.0 Testversion
//             - Der Quelltext der Firmware wurde in der Programmiersprache
//               C neu geschrieben
//             09.09.2003 - V2.0
//             - Fehler in der Generierung des DMX-Signals behoben
//             17.05.2006 - V2.1
//             - Kufer: Anpassung auf OpenDCC-Hardware
//               (Ports, Clock 9.216->16.000MHz, Uart)
//
// Copyright (C) 2001,2002,2003 Mathias Dzionsko (madz@gmx.de)
//===========================================================================
#include <inttypes.h>
#include <avr\io.h>
#include <avr\interrupt.h>
#include <util\delay.h>
//---------------------------------------------------------------------------
// Makros
// wie in delay.h, jedoch war dort der Bereich zu klein!
static inline void _mydelay_us(double __us) __attribute__((always_inline));
void
_mydelay_us(double __us)
{
	uint16_t __ticks;
	double __tmp = ((F_CPU) / 4e6) * __us;
	if (__tmp < 1.0)
		__ticks = 1;
	else if (__tmp > 65535)
		__ticks = 0;	/* i.e. 65536 */
	else
		__ticks = (uint16_t)__tmp;
	_delay_loop_2(__ticks);
}


//---------------------------------------------------------------------------
// Konstanten

#define BAUDRATE 115200L

// DMX
#define DMX_PORT        PORTB
#define DMX_PIN         PB4

// LED_RS232 zeigt ein erfolgreiches Senden der DMX Daten an -> LED1
// LED_STOP zeigt Fehler beim Datenempfang vom PC an -> LED2
// LED_GO zeigt das Ende das Startvorgangs an (Daten knnen nun
// gesendet werden) -> LED3

// Leuchtdauer der LEDs: 100ms (in 1/100s)
#define LED1_ONTIME     10
#define LED2_ONTIME     10

//------------------------------------------------------------------------
// Port Definitions (aus config.h)
//------------------------------------------------------------------------
// Port A
#define S88DIN1    0     // in
#define S88DIN2    1     // in
#define S88DIN3    2     // in
#define EXT_STOP   3     // in
#define LED_GO     4     // out
#define LED_STOP   5     // out
#define KEY_STOP   6     // in
#define KEY_GO     7     // in
// Port B
#define S88_CLK    0     // out
#define S88_RESET  1     // out
#define ACK_DETECT 2     // in    this is only a short pulse -> use int!
#define S88_LOAD   3     // out
#define DMX_OUT    4     // out
#define MY_CTS     6     // out
#define MY_RTS     7     // in
// Port C
#define SHORT_MAIN 0     // in
#define SHORT_PROG 1     // in
#define LED_CTRL   6     // out
#define LED_RS232  7     // out
// Port D
#define MY_RXD     0     // in
#define MY_TXD     1     // out
#define JUMPER1    2     // in
#define JUMPER2    3     // in
#define NDCC       4     // out
#define DCC        5     // out
#define ENA_PROG   6     // out  high = enable programming
#define ENA_MAIN   7     // out  high = enable main

#define LED_CTRL_ON      PORTC &= ~(1<<LED_CTRL)
#define LED_CTRL_OFF     PORTC |= (1<<LED_CTRL)
#define LED_RS232_ON     PORTC &= ~(1<<LED_RS232)
#define LED_RS232_OFF    PORTC |= (1<<LED_RS232)
#define LED_GO_ON        PORTA &= ~(1<<LED_GO)
#define LED_GO_OFF       PORTA |= (1<<LED_GO)
#define LED_STOP_ON      PORTA &= ~(1<<LED_STOP)
#define LED_STOP_OFF     PORTA |= (1<<LED_STOP)



// UART (Werte fr 9216MHz)
#define UBRR_INIT       4                       // 115200 Baud
#define UCR_INIT        (_BV(RXCIE)+_BV(RXEN)+_BV(TXEN))


#define RX_BUFSIZE      64
#define RX_TIMEOUT      10 // 100ms (in 1/100s)

// Timer
#define TIMSK_INIT      _BV(TOIE0)
#define TCCR0_INIT      (_BV(CS02)+_BV(CS00))   // CK/1024 => 16000 Hz // war: 9000 Hz
#define TIMER0_RELOAD   (256-160)                // => 100 Hz  // war: 256-90

// Kommunikationsprotokoll
#define P_BLKSTART      0x5A
#define P_BLKEND        0xA5

#define P_DMXOUT96      0xA0
#define P_DMXOUT256     0xA1
#define P_DMXOUT512     0xA2

#define P_ERROR         0xC0
#define P_ACK           0xC1

// Konstanten fr die Variable Error
#define E_BUFFEROVERFLOW        1
#define E_UARTTIMEOUT           2
#define E_PROTOCOLERROR         4
//---------------------------------------------------------------------------
// Globale Variablen

volatile uint8_t        TimerLED1, TimerLED2;   // LED Timer
volatile uint8_t        UARTTimer;
volatile uint8_t        Error;

// Empgangspuffer fr die serielle Schnittstelle
uint8_t                 RxBuffer[RX_BUFSIZE];
uint8_t                 RxBufferOut;
volatile uint8_t        RxBufferIn;
//---------------------------------------------------------------------------
// OVERFLOW0 Interrupt

ISR(SIG_OVERFLOW0)
{
  // Timer nachladen
  TCNT0=TIMER0_RELOAD;

  // UARTTimer
  if (UARTTimer) UARTTimer--;
  
  // LED1 ein/ausschalten
  if (TimerLED1)
    {
      LED_RS232_ON;
      TimerLED1--;
    }
  else 
      LED_RS232_OFF;
  

  // LED2 ein/ausschalten
    if (TimerLED2)
	  {
        LED_STOP_ON;
	    TimerLED2--;
      }
	else LED_STOP_OFF;
  }
//---------------------------------------------------------------------------
// SIG_UART_RECV Interrupt

ISR(SIG_UART_RECV)
{
  uint8_t nextpos, dummy;

  // Nchste Bufferposition ausrechnen
  nextpos=RxBufferIn+1;
  if (nextpos==RX_BUFSIZE) nextpos=0;

  // Ist der Buffer voll?
  if (nextpos==RxBufferOut) {
    // Buffer ist bergelaufen
    Error|=E_BUFFEROVERFLOW;
    dummy=UDR; // UART Data Register leeren
  } else {
    // Empfangenes Byte in den Buffer schreiben
    RxBuffer[RxBufferIn]=UDR;
    RxBufferIn=nextpos;
  }
}
//---------------------------------------------------------------------------
// UARTRecvByte

uint8_t UARTRecvByte(void)
{
  uint8_t data;

  // Warten bis ein Byte im Buffer oder die Zeit abgelaufen ist
  UARTTimer=RX_TIMEOUT;
  while (RxBufferIn==RxBufferOut && UARTTimer) ;

  // Kein Byte angekommen?
  if (RxBufferIn==RxBufferOut) {
    Error|=E_UARTTIMEOUT;
    return 0;
  }
  
  // Byte auslesen
  data=RxBuffer[RxBufferOut++];
  if (RxBufferOut==RX_BUFSIZE) RxBufferOut=0;

  return data;
}
//---------------------------------------------------------------------------
// UARTRSendByte

void UARTSendByte(uint8_t data)
{
  // Warten bis das letzte Byte gesendet wurde
  loop_until_bit_is_set(UCSRA, UDRE);

  // Byte senden
  UDR=data;
}
//---------------------------------------------------------------------------
// DMXSendByte

// Sendet ein Datenbyte ber die DMX-Leitung.
// Die bentigten Takte fr ein DMX-Bit sind 37. Die Taktfrequenz betrgt
// 9,216 MHz. Das ergibt 4,0147 s pro Bit. Die Abweichung betrgt dann 0,37
// Prozent, welche innerhalb der zulssigen Toleranz liegt.
// Interrupts werden fr 44,3 s nicht zugelassen.

// Kufer: neu: 16MHz, es sind 64 Takte ntig, entsprechend die
// Wartezeit verlngert.

void DMXSendByte(uint8_t value)
{
  asm volatile (
"                cli                    \n" // keine Ints zulassen
"                clc                    \n" // 1      - Start Bit
"                rcall DMXSendBit       \n" // +3+33=37 Takte
"                                       \n"
"                ror %2                 \n" // 1      - Bit 0
"                rcall DMXSendBit       \n" // +3+33=37 Takte
"                ror %2                 \n" //        - Bit 1
"                rcall DMXSendBit       \n"
"                ror %2                 \n" //        - Bit 2
"                rcall DMXSendBit       \n"
"                ror %2                 \n" //        - Bit 3
"                rcall DMXSendBit       \n"
"                ror %2                 \n" //        - Bit 4
"                rcall DMXSendBit       \n"
"                ror %2                 \n" //        - Bit 5
"                rcall DMXSendBit       \n"
"                ror %2                 \n" //        - Bit 6
"                rcall DMXSendBit       \n"
"                ror %2                 \n" //        - Bit 7
"                rcall DMXSendBit       \n"
"                                       \n"
"                sec                    \n"  //        - Stop Bit 1
"                rcall DMXSendBit       \n"
"                sec                    \n"  //        - Stop Bit 2
"                rcall DMXSendBit       \n"
"                sei                    \n"  // Ints wieder zulassen
"                rjmp DMXSendByteExit   \n"
"                                       \n"
"DMXSendBit:     brcs DMXSendBit1       \n"
"                nop                    \n" // 1+1
"                cbi %0, %1             \n" // +2
"                rjmp DMXSendBit2       \n" // +2=6 Takte
"DMXSendBit1:    sbi %0, %1             \n" // 2+2
"                rjmp DMXSendBit2       \n" // +2=6 Takte
"                                       \n"
"DMXSendBit2:    ldi r31,15             \n" // 1  : Kufer: war 6, jetzt 15
"DMXSendBit3:    dec r31                \n" // +1
"                brne DMXSendBit3       \n" // +1/+2
"                rjmp DMXSendBit4       \n" // +2
"DMXSendBit4:    ret                    \n" // +4=9+6*3=27 Takte
"DMXSendByteExit:                       \n" // neu: 54 Takte
//                          %0               %1            %2
    : : "I" (_SFR_IO_ADDR(DMX_PORT)), "I" (DMX_PIN), "r" (value) : "r31"
  );
}
//---------------------------------------------------------------------------
// DMXSendReset

// Sendet ein Reset-Signal mit anschlieendem Start-Byte ber die
// DMX-Leitung.

void DMXSendReset(void)
  {
    DMX_PORT &= ~(1<<DMX_PIN);          // RESET-Signal senden
    _mydelay_us(100);                  // 100s (min. 88s) warten

    DMX_PORT |= (1<<DMX_PIN);             // MARK-Signal senden
    _mydelay_us(10);                   // 10s (min. 8s) warten

    DMXSendByte(0);                  // Startbyte (0) senden
  }
//---------------------------------------------------------------------------
// DMXSendChannels

void DMXSendChannels(uint16_t cnt)
{
  uint8_t data;

  DMXSendReset();
  
  while (cnt--) {
    data=UARTRecvByte();
    if (Error) return;
    DMXSendByte(data);
  }
}
//---------------------------------------------------------------------------
// Hauptfunktion

int main(void)
{
  uint8_t data;

  // Port B initialisieren
  // PORTB=PORTB_INIT;
  // DDRB=DDRB_INIT;

  // Port D initialisieren
  // PORTD=PORTD_INIT;
  // DDRD=DDRD_INIT;
  
// Input/Output Ports initialization
  PORTA = (1<<S88DIN1)       // high = pull up
        | (1<<S88DIN2)       // high = pull up
        | (1<<S88DIN3)       // high = pull up
        | (1<<EXT_STOP)      // high = pull up
        | (1<<LED_GO)        // low = turn on
        | (1<<LED_STOP)      // low = turn on
        | (1<<KEY_STOP)      // high = pull up
        | (1<<KEY_GO);       // high = pull up
  DDRA  = (0<<S88DIN1)       // in
        | (0<<S88DIN2)       // in
        | (0<<S88DIN3)       // in
        | (0<<EXT_STOP)      // in
        | (1<<LED_GO)        // out
        | (1<<LED_STOP)      // out
        | (0<<KEY_STOP)      // in
        | (0<<KEY_GO);       // in
  // Port B
  PORTB = (0<<S88_CLK)       // clk = low
        | (0<<S88_RESET)     // reset = low
        | (1<<ACK_DETECT)    // pullup
        | (1<<S88_LOAD)      // load = high
        | (1<<DMX_OUT)       // high = inactiv
		| (0<<MY_CTS)
        | (1<<MY_RTS);       // high = pullup

  DDRB  = (1<<S88_CLK)       // out
        | (1<<S88_RESET)     // out
        | (0<<ACK_DETECT)    // in
        | (1<<S88_LOAD)      // out
        | (1<<DMX_OUT)       // out
        | (1<<MY_CTS)        // out
        | (0<<MY_RTS);       // in
// Port C
  PORTC = (1<<SHORT_MAIN)    // pull up
        | (1<<SHORT_PROG)    // pull up
        | (1<<LED_CTRL)      // low = turn on
        | (1<<LED_RS232);    // low = turn on

  DDRC  = (0<<SHORT_MAIN)    // in
        | (0<<SHORT_PROG)    // in
        | (1<<LED_CTRL)      // out
        | (1<<LED_RS232);    // out
// Port D
  PORTD = (1<<MY_RXD)        // in, pullup
        | (1<<MY_TXD)        // out
        | (1<<JUMPER1)       // pullup
        | (1<<JUMPER2)       // pullup
        | (0<<NDCC)          // out
        | (1<<DCC)           // out
        | (0<<ENA_PROG)      // out, disabled
        | (0<<ENA_MAIN);     // out, disabled

  DDRD  = (0<<MY_RXD)        // in
        | (1<<MY_TXD)        // out
        | (0<<JUMPER1)       // in
        | (0<<JUMPER2)       // in
        | (1<<NDCC)          // out
        | (1<<DCC)           // out
        | (1<<ENA_PROG)      // out
        | (1<<ENA_MAIN);     // out

  // UART initialisieren

    uint16_t ubrr = (uint16_t) ((uint32_t) F_CPU/(8*BAUDRATE) - 1);
     
    UBRRH = (uint8_t) (ubrr>>8);
    UBRRL = (uint8_t) (ubrr);
//   UBRRL = 0xCF; // 207: 4800bps @ 16.00MHz
//   UBRRL = 103; // 9600bps @ 16.00MHz
//   UBRRL = 51; // 19200bps @ 16.00MHz
//   UBRRL = 25; // 38400bps @ 16.00MHz
   
    // Rcksetzen von Receive und Transmit Complete-Flags
    // UART Receiver und Transmitter anschalten, Receive-Interrupt aktivieren
    // Data mode 8N1, asynchron, aber Highspeed, wegen hoher Baudrate (nur div 8)
    UCSRA = (1 << RXC) | (1 << TXC) | (1 << U2X);
    UCSRB = (1 << RXEN) | (1 << TXEN) | (1 << RXCIE);
    UCSRC = (1 << URSEL) | (1 << UCSZ1) | (1 << UCSZ0);

  // LED-Timer initialisieren
  TCNT0=TIMER0_RELOAD;
  TCCR0=TCCR0_INIT;
  TIMSK=TIMSK_INIT;

  // Interrupts zulassen
  sei();

  // LED3 einschalten
  LED_GO_ON;
  
  // Hauptschleife
  while (1) {
    Error=0;
    data=UARTRecvByte();
    if (!Error) {
      // Blockanfang?
      if (data==P_BLKSTART) {
        data=UARTRecvByte();
        if (!Error) {
          // Befehl vom PC ausfhren
          switch (data) {
            case P_DMXOUT96: DMXSendChannels(96); break;
            case P_DMXOUT256: DMXSendChannels(256); break;
            case P_DMXOUT512: DMXSendChannels(512); break;
            default: Error|=E_PROTOCOLERROR;
          }
          if (!Error) {
            // Stimmt das Blockende?
            data=UARTRecvByte();
            if (data!=P_BLKEND) Error|=E_PROTOCOLERROR;
          }
        }
      } else Error|=E_PROTOCOLERROR;
    }
    if (Error) {
      // Fehler anzeigen
      TimerLED2=LED2_ONTIME;
      // Fehler an den PC senden
      UARTSendByte(P_BLKSTART);
      UARTSendByte(P_ERROR);
      UARTSendByte(P_BLKEND);
    } else {
      // Befehl erfolgreich bearbeitet
      TimerLED1=LED1_ONTIME;
      // Besttigung an den PC senden
      UARTSendByte(P_BLKSTART);
      UARTSendByte(P_ACK);
      UARTSendByte(P_BLKEND);
    }
  }
  return 0;
}
//---------------------------------------------------------------------------
